/* This file was originally... */
/*
 *	$Id: pci.c,v 1.8 2006/01/06 21:19:48 josejx Exp $
 *
 *	Linux PCI Utilities -- List All PCI Devices
 *
 *	Copyright (c) 1997--1999 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */
/* So, all the same terms apply but now it's also:
 * Copyright (c) 2000 Tom Rini <trini@kernel.crashing.org>
 */

#include <string.h>
#include <stdlib.h>
#include <pci/pci.h>

struct pci_filter filter;	/* Device filter */
struct pci_access *pacc;
int bus, dev, func; /* Location of the card */

struct device {
  struct device *next;
  struct pci_dev *dev;
  unsigned int config_cnt;
  u8 config[256];
};

struct device *first_dev;

struct device *scan_device(struct pci_dev *p)
{
  int how_much = 64;
  struct device *d;

  if (!pci_filter_match(&filter, p))
    return NULL;
  d = malloc(sizeof(struct device));
  bzero(d, sizeof(*d));
  d->dev = p;
  if (!pci_read_block(p, 0, d->config, how_much))
    exit(1);
  if (how_much < 128 && (d->config[PCI_HEADER_TYPE] & 0x7f) == PCI_HEADER_TYPE_CARDBUS)
    {
      /* For cardbus bridges, we need to fetch 64 bytes more to get the full standard header... */
      if (!pci_read_block(p, 64, d->config+64, 64))
	exit(1);
      how_much = 128;
    }
  d->config_cnt = how_much;
  pci_setup_cache(p, d->config, d->config_cnt);
  pci_fill_info(p, PCI_FILL_IDENT | PCI_FILL_IRQ | PCI_FILL_BASES | PCI_FILL_ROM_BASE | PCI_FILL_SIZES);
  return d;
}
void scan_devices(void)
{
  struct device *d;
  struct pci_dev *p;

  pci_scan_bus(pacc);
  for(p=pacc->devices; p; p=p->next)
    if ((d = scan_device(p)))
      {
	d->next = first_dev;
	first_dev = d;
      }
}

u16 get_conf_word(struct device *d, unsigned int pos)
{
  return d->config[pos] | (d->config[pos+1] << 8);
}

void findVideoDevice(void)
{
  struct device *d;
  struct pci_dev *p;
  u16 w;
  
  pacc = pci_alloc();
  pci_filter_init(pacc, &filter);
  pci_init(pacc);
  scan_devices();

  for(d=first_dev; d; d=d->next) {
    p = d->dev;
    /* See if the current device is a VGA card, if so set vars and break */
    w = get_conf_word(d, PCI_CLASS_DEVICE) >> 8;
    if(w == PCI_BASE_CLASS_DISPLAY) {
      bus = p->bus;
      dev = p->dev;
      func = p->func;
      break;
    } 
  }

  pci_cleanup(pacc);
}
